#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_jwt.zmi.plugin module

This module is used to register ZMI views used to manage JWT configuration.
"""

from pyams_auth_jwt.interfaces import IJWTSecurityConfiguration, JWT_PROXY_CACHE_NAME, \
    JWT_PROXY_TOKENS_NAMESPACE
from pyams_form.ajax import ajax_form_config
from pyams_form.browser.checkbox import SingleCheckBoxFieldWidget
from pyams_form.field import Fields
from pyams_form.interfaces.form import IGroup
from pyams_layer.interfaces import IPyAMSLayer
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.base import MANAGE_SECURITY_PERMISSION
from pyams_security_views.zmi import ISecurityMenu
from pyams_skin.interfaces.viewlet import IHeaderViewletManager
from pyams_skin.viewlet.help import AlertMessage
from pyams_utils.adapter import adapter_config
from pyams_utils.cache import clear_cache
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminEditForm, FormGroupChecker
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem


__docformat__ = 'restructuredtext'

from pyams_auth_jwt import _  # pylint: disable=ungrouped-imports


@viewlet_config(name='jwt-security-configuration.menu',
                context=ISecurityManager, layer=IAdminLayer,
                manager=ISecurityMenu, weight=50,
                permission=MANAGE_SECURITY_PERMISSION)
class JWTSecurityConfiguration(NavigationMenuItem):
    """JWT security configuration menu"""

    label = _("JWT configuration")
    href = '#jwt-security-configuration.html'


@ajax_form_config(name='jwt-security-configuration.html',
                  context=ISecurityManager, layer=IPyAMSLayer,
                  permission=MANAGE_SECURITY_PERMISSION)
class JWTSecurityConfigurationEditForm(AdminEditForm):
    """JWT security configuration edit form"""

    title = _("JWT configuration")
    legend = _("JWT tokens settings")

    fields = Fields(IJWTSecurityConfiguration).select('access_token_name', 'refresh_token_name')

    def get_content(self):
        return IJWTSecurityConfiguration(self.context)

    def apply_changes(self, data):
        configuration = self.get_content()
        old_region = configuration.selected_cache
        changes = super(JWTSecurityConfigurationEditForm, self).apply_changes(data)
        if changes and (old_region is not None):
            clear_cache(JWT_PROXY_CACHE_NAME, old_region, JWT_PROXY_TOKENS_NAMESPACE)
        return changes


@adapter_config(name='jwt-configuration',
                required=(ISecurityManager, IAdminLayer, JWTSecurityConfigurationEditForm),
                provides=IGroup)
class JWTConfigurationGroup(FormGroupChecker):
    """JWT configuration edit group"""

    fields = Fields(IJWTSecurityConfiguration).select('local_mode', 'algorithm', 'secret',
                                                      'private_key', 'public_key',
                                                      'access_expiration', 'refresh_expiration')
    weight = 10

    def update_widgets(self, prefix=None, use_form_mode=True):
        super().update_widgets(prefix, use_form_mode)
        widget = self.widgets.get('private_key')
        if widget is not None:
            widget.prefix = JWTConfigurationKeyAlert(self.context, self.request, self, None)
            widget.rows = 15
            widget.add_class('monospace')
        widget = self.widgets.get('public_key')
        if widget is not None:
            widget.rows = 5
            widget.add_class('monospace')


@viewlet_config(name='jwt-configuration.header',
                context=ISecurityManager, layer=IAdminLayer, view=JWTConfigurationGroup,
                manager=IHeaderViewletManager, weight=1)
class JWTConfigurationHeader(AlertMessage):
    """JWT configuration header"""

    status = 'info'

    _message = _("JWT authentication module \"local mode\" allows to generate, check and refresh "
                 "tokens locally.\n"
                 "You can choose to use a simple secret key to encrypt your tokens, or to "
                 "use a private and public keys pair (which can to be used to share tokens "
                 "between two applications).")


class JWTConfigurationKeyAlert(AlertMessage):
    """JWT configuration keys alert"""

    status = 'info'
    css_class = 'mb-1 p-2'

    _message = _("""You can use the `openssl` command to generate your keys:

    openssl genpkey -algorithm RSA -out private-key.pem
    openssl rsa -pubout -in private-key.pem -out public-key.pem
""")
    message_renderer = 'markdown'


@adapter_config(name='jwt-proxy-configuration',
                required=(ISecurityManager, IAdminLayer, JWTSecurityConfigurationEditForm),
                provides=IGroup)
class JWTProxyConfigurationGroup(FormGroupChecker):
    """JWT proxy configuration edit group"""

    fields = Fields(IJWTSecurityConfiguration).select(
        'proxy_mode', 'authority', 'get_token_service', 'proxy_access_token_name',
        'proxy_refresh_token_name', 'get_claims_service', 'refresh_token_service',
        'verify_token_service', 'verify_ssl')
    fields['verify_ssl'].widget_factory = SingleCheckBoxFieldWidget

    weight = 20

    def get_content(self):
        return IJWTSecurityConfiguration(self.context)


@viewlet_config(name='jwt-proxy-configuration.header',
                context=ISecurityManager, layer=IAdminLayer, view=JWTProxyConfigurationGroup,
                manager=IHeaderViewletManager, weight=1)
class JWTProxyConfigurationHeader(AlertMessage):
    """JWT proxy configuration header"""

    status = 'info'

    _message = _("JWT authentication module \"proxy mode\" relies on another authentication "
                 "authority (which can be another application using this JWT package) to "
                 "generate, check and refresh tokens. This authority can be used to share "
                 "access tokens between different applications.\n"
                 "You can cache tokens to reduce the number of requests which will be forwarded "
                 "to the authentication authority.")


@adapter_config(name='jwt-proxy-cache-configuration',
                required=(ISecurityManager, IAdminLayer, JWTProxyConfigurationGroup),
                provides=IGroup)
class JWTProxyCacheConfigurationGroup(FormGroupChecker):
    """JWT proxy cache configuration edit group"""

    fields = Fields(IJWTSecurityConfiguration).select('use_cache', 'selected_cache')



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_security.plugin.jwt module

This module provides a JWT authentication plug-in.
"""

import logging
from datetime import datetime, timedelta, timezone

import jwt
from ZODB.POSException import ConnectionStateError
from jwt import InvalidTokenError
from persistent import Persistent
from pyramid.httpexceptions import HTTPUnauthorized
from zope.container.contained import Contained
from zope.schema.fieldproperty import FieldProperty

from pyams_auth_jwt.interfaces import ACCESS_OBJECT, IJWTAuthenticationPlugin, \
    IJWTProxyHandler, IJWTSecurityConfiguration, JWT_CONFIGURATION_KEY
from pyams_security.credential import Credentials
from pyams_security.interfaces import ISecurityManager
from pyams_security.interfaces.plugin import ICredentialsPlugin
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.property import ClassPropertyType, classproperty
from pyams_utils.registry import get_current_registry, query_utility, utility_config


__docformat__ = 'restructuredtext'

from pyams_auth_jwt import _  # pylint: disable=ungrouped-imports


LOGGER = logging.getLogger('PyAMS (JWT security)')

PARSED_CLAIMS_ENVKEY = 'pyams_auth_jwt.claims'
PARSED_CREDENTIALS_ENVKEY = "pyams_auth_jwt.credentials"


@factory_config(IJWTSecurityConfiguration)
class JWTSecurityConfiguration(Persistent, Contained):
    """JWT security configuration"""

    local_mode = FieldProperty(IJWTSecurityConfiguration['local_mode'])
    algorithm = FieldProperty(IJWTSecurityConfiguration['algorithm'])
    secret = FieldProperty(IJWTSecurityConfiguration['secret'])
    private_key = FieldProperty(IJWTSecurityConfiguration['private_key'])
    public_key = FieldProperty(IJWTSecurityConfiguration['public_key'])
    access_expiration = FieldProperty(IJWTSecurityConfiguration['access_expiration'])
    access_token_name = FieldProperty(IJWTSecurityConfiguration['access_token_name'])
    refresh_expiration = FieldProperty(IJWTSecurityConfiguration['refresh_expiration'])
    refresh_token_name = FieldProperty(IJWTSecurityConfiguration['refresh_token_name'])

    proxy_mode = FieldProperty(IJWTSecurityConfiguration['proxy_mode'])
    authority = FieldProperty(IJWTSecurityConfiguration['authority'])
    get_token_service = FieldProperty(IJWTSecurityConfiguration['get_token_service'])
    proxy_access_token_name = FieldProperty(IJWTSecurityConfiguration['proxy_access_token_name'])
    get_claims_service = FieldProperty(IJWTSecurityConfiguration['get_claims_service'])
    refresh_token_service = FieldProperty(IJWTSecurityConfiguration['refresh_token_service'])
    proxy_refresh_token_name = FieldProperty(IJWTSecurityConfiguration['proxy_refresh_token_name'])
    verify_token_service = FieldProperty(IJWTSecurityConfiguration['verify_token_service'])
    verify_ssl = FieldProperty(IJWTSecurityConfiguration['verify_ssl'])
    use_cache = FieldProperty(IJWTSecurityConfiguration['use_cache'])
    selected_cache = FieldProperty(IJWTSecurityConfiguration['selected_cache'])

    @property
    def enabled(self):
        """Check if configuration is enabled"""
        return self.local_mode or self.proxy_mode


@adapter_config(required=ISecurityManager, provides=IJWTSecurityConfiguration)
def securiy_manager_jwt_configuration_factory(context):
    """Security manager JWT configuration factory adapter"""
    return get_annotation_adapter(context, JWT_CONFIGURATION_KEY, IJWTSecurityConfiguration)


@utility_config(provides=IJWTAuthenticationPlugin)
@utility_config(name='jwt', provides=ICredentialsPlugin)
class JWTAuthenticationPlugin(metaclass=ClassPropertyType):
    """JWT authentication plugin"""

    prefix = 'jwt'
    title = _("JWT authentication")

    audience = None
    leeway = 0
    callback = None
    json_encoder = None

    @classproperty
    def http_header(cls):  # pylint: disable=no-self-argument,no-self-use
        """HTTP header setting"""
        return get_current_registry().settings.get('pyams_auth_jwt.http_header', 'Authorization')

    @classproperty
    def auth_type(cls):  # pylint: disable=no-self-argument,no-self-use
        """HTTP authentication type setting"""
        return get_current_registry().settings.get('pyams_auth_jwt.auth_type', 'Bearer')

    @property
    def configuration(self):  # pylint: disable=no-self-use
        """JWT configuration getter"""
        try:
            manager = query_utility(ISecurityManager)
            if manager is not None:
                return IJWTSecurityConfiguration(manager)
        except ConnectionStateError:
            return None
        return None

    @property
    def enabled(self):
        """Check if JWT authentication is enabled in security manager"""
        configuration = self.configuration
        # pylint: disable=no-member
        try:
            return configuration.enabled if (configuration is not None) else False
        except ConnectionStateError:
            return False

    @property
    def expiration(self):
        """Get default security manager expiration"""
        configuration = self.configuration
        # pylint: disable=no-member
        return configuration.access_expiration if configuration is not None else None

    def create_token(self, principal, expiration=None, audience=None, **claims):
        """Create JWT token"""
        if not self.enabled:
            return None
        configuration = self.configuration
        payload = {}
        payload.update(claims)
        payload['sub'] = principal
        payload['iat'] = iat = datetime.now(timezone.utc)
        expiration = expiration or self.expiration
        if expiration:
            if not isinstance(expiration, timedelta):
                expiration = timedelta(seconds=expiration)
            payload['exp'] = iat + expiration
        audience = audience or self.audience
        if audience:
            payload['aud'] = audience
        # pylint: disable=no-member
        algorithm = configuration.algorithm if configuration is not None else 'RS512'
        if algorithm.startswith('HS'):
            # pylint: disable=no-member
            key = configuration.secret if configuration is not None else None
        else:  # RS256
            # pylint: disable=no-member
            key = configuration.private_key if configuration is not None else None
        token = jwt.encode(payload, key, algorithm=algorithm, json_encoder=self.json_encoder)
        if not isinstance(token, str):
            token = token.decode('ascii')
        return token

    def _get_claims(self, request, obj=None):  # pylint: disable=too-many-return-statements
        """Get JWT claims"""
        if self.http_header == 'Authorization':  # pylint: disable=comparison-with-callable
            try:
                if request.authorization is None:
                    return {}
            except (ValueError, AttributeError):  # invalid authorization header
                return {}
            (auth_type, token) = request.authorization
            if auth_type != self.auth_type:  # pylint: disable=comparison-with-callable
                return {}
        else:
            token = request.headers.get(self.http_header)
        if not token:
            return {}
        try:
            configuration = self.configuration
            # pylint: disable=no-member
            algorithm = configuration.algorithm if configuration is not None else 'RS512'
            if algorithm.startswith('HS'):
                # pylint: disable=no-member
                key = configuration.secret if configuration is not None else None
            else:  # RS256/RS512
                # pylint: disable=no-member
                key = configuration.public_key if configuration is not None else None
            claims = jwt.decode(token, key, algorithms=[algorithm],
                                leeway=self.leeway, audience=self.audience)
            if obj and obj != claims.get('obj'):
                raise InvalidTokenError('Bad token object!')
            return claims
        except InvalidTokenError as exc:
            LOGGER.warning('Invalid JWT token from %s: %s',
                           getattr(request, 'remote_addr', '--'), exc)
            return {}

    def get_claims(self, request, obj=None):  # pylint: disable=too-many-return-statements
        """Get JWT claims"""
        configuration = self.configuration
        if configuration is None:
            return {}
        if configuration.proxy_mode:
            handler = IJWTProxyHandler(self, None)
            if handler is not None:
                _status_code, claims = handler.get_claims(request, obj)  # pylint: disable=assignment-from-no-return
                return claims
        elif configuration.local_mode:
            return self._get_claims(request, obj)
        return {}

    def extract_credentials(self, request, **kwargs):  # pylint: disable=unused-argument
        """Extract principal ID from given request"""
        claims = self.get_claims(request, obj=ACCESS_OBJECT)
        if claims:
            return Credentials(self.prefix,
                               claims.get('sub'),
                               login=claims.get('login')) if claims else None
        return None

    def authenticate(self, credentials, request):  # pylint: disable=unused-argument
        """Authenticate JWT token"""
        claims = self.get_claims(request, obj=ACCESS_OBJECT)
        return claims.get('sub') if claims else None


@adapter_config(required=IJWTAuthenticationPlugin,
                provides=IJWTSecurityConfiguration)
def jwt_plugin_configuration_adapter(context):
    """JWT plugin configuration adapter"""
    return context.configuration


def create_jwt_token(request, principal, expiration=None, audience=None, **claims):
    # pylint: disable=unused-argument
    """Create JWT token"""
    plugin = query_utility(IJWTAuthenticationPlugin)
    if (plugin is not None) and plugin.enabled:
        return plugin.create_token(principal, expiration, audience, **claims)
    return None


def get_jwt_claims(request, obj=None):
    """Get JWT claims"""
    plugin = query_utility(IJWTAuthenticationPlugin)
    if (plugin is not None) and plugin.enabled:
        return plugin.get_claims(request, obj=obj)
    return {}


def jwt_object_view(view, info):
    """JWT object view deriver
    
    This filter is raising an HTTP Unauthorized exception if required JWT
    token object is not provided by the incoming request
    """
    jwt_object = info.options.get('jwt_object')
    if jwt_object:
        def view_wrapper(context, request):
            claims = get_jwt_claims(request, obj=jwt_object)
            if not claims:
                raise HTTPUnauthorized
            return view(context, request)
        return view_wrapper
    return view
    
jwt_object_view.options = ('jwt_object',)



#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_jwt.api module

This modules is used to register REST API to generate, check and refresh JWT tokens.
"""

import sys

from colander import Int, MappingSchema, OneOf, SchemaNode, String, drop
from cornice import Service
from cornice.validators import colander_body_validator
from pyramid.authorization import Authenticated
from pyramid.httpexceptions import HTTPAccepted, HTTPBadRequest, HTTPForbidden, HTTPOk, HTTPServiceUnavailable, \
    HTTPUnauthorized

from pyams_auth_jwt.interfaces import ACCESS_OBJECT, IJWTProxyHandler, \
    IJWTSecurityConfiguration, REFRESH_OBJECT, REST_TOKEN_ROUTE, REST_VERIFY_ROUTE
from pyams_auth_jwt.plugin import create_jwt_token, get_jwt_claims as get_request_claims
from pyams_security.credential import Credentials
from pyams_security.interfaces import ISecurityManager
from pyams_security.rest import check_cors_origin, set_cors_headers
from pyams_utils.registry import query_utility
from pyams_utils.rest import BaseStatusSchema, PropertiesMapping, STATUS, http_error, rest_responses

__docformat__ = 'restructuredtext'


TEST_MODE = sys.argv[-1].endswith('/test')


class ClaimsInfo(MappingSchema):
    """Claims setter schema"""
    claims = SchemaNode(PropertiesMapping(),
                        description="Custom claims",
                        missing=drop)


class LoginInfo(ClaimsInfo):
    """Login schema"""
    login = SchemaNode(String(),
                       description="User login")
    password = SchemaNode(String(),
                          description="User password")


class TokensResult(BaseStatusSchema):
    """Tokens result schema"""
    accessToken = SchemaNode(String(),
                             description="Access token")
    refreshToken = SchemaNode(String(),
                              description="Refresh token")


class ClaimsObjectInfo(MappingSchema):
    """Claims object info"""
    obj = SchemaNode(String(),
                     description="Token object",
                     validator=OneOf((ACCESS_OBJECT, REFRESH_OBJECT)),
                     missing=drop)


class ClaimsElements(ClaimsObjectInfo):
    """Token claims elements schema"""
    sub = SchemaNode(String(),
                     description="Principal ID")
    iat = SchemaNode(Int(),
                     description="Token issue timestamp, in seconds")
    exp = SchemaNode(Int(),
                     description="Token expiration timestamp, in seconds")


#
# JWT token management service
#

jwt_token = Service(name=REST_TOKEN_ROUTE,
                    pyramid_route=REST_TOKEN_ROUTE,
                    description="JWT tokens management")


@jwt_token.options(validators=(check_cors_origin, set_cors_headers))
def jwt_token_options(request):  # pylint: disable=unused-argument
    """JWT token OPTIONS handler"""
    return ''


class JWTTokenPostResponse(MappingSchema):
    """Token post response"""

    body = TokensResult()


jwt_token_post_responses = rest_responses.copy()
jwt_token_post_responses[HTTPOk.code] = JWTTokenPostResponse(
    description="Get new access and refresh tokens matching credentials")


@jwt_token.post(require_csrf=False,
                content_type=('application/json', 'multipart/form-data'),
                schema=LoginInfo(),
                validators=(check_cors_origin, colander_body_validator, set_cors_headers),
                response_schemas=jwt_token_post_responses)
def get_jwt_token(request):
    """Get new access and refresh tokens matching credentials"""
    # check security manager utility
    sm = query_utility(ISecurityManager)  # pylint: disable=invalid-name
    if sm is None:
        return http_error(request, HTTPServiceUnavailable)
    configuration = IJWTSecurityConfiguration(sm)
    if not configuration.enabled:
        return http_error(request, HTTPServiceUnavailable)
    # check request params
    params = request.params if TEST_MODE else request.validated
    login = params.get('login')
    if not login:
        return http_error(request, HTTPBadRequest, 'missing credentials')
    credentials = Credentials('jwt', id=login, **params)
    # use remote authentication authority
    if configuration.proxy_mode:
        handler = IJWTProxyHandler(sm, None)
        if handler is not None:
            status_code, tokens = handler.get_tokens(request, credentials)  # pylint: disable=assignment-from-no-return
            request.response.status_code = status_code
            return tokens
    # authenticate principal in security manager
    principal_id = sm.authenticate(credentials, request)
    if principal_id is None:
        return http_error(request, HTTPUnauthorized, "invalid credentials")
    custom_claims = params.get('claims', {})
    request.response.cache_expires(configuration.refresh_expiration)
    return {
        'status': STATUS.SUCCESS.value,
        configuration.access_token_name:
            create_jwt_token(request,
                             principal_id,
                             expiration=configuration.access_expiration,
                             obj=ACCESS_OBJECT,
                             **custom_claims),
        configuration.refresh_token_name:
            create_jwt_token(request,
                             principal_id,
                             expiration=configuration.refresh_expiration,
                             obj=REFRESH_OBJECT)
    }


class JWTTokenGetResponse(MappingSchema):
    """Token claims getter response"""

    body = ClaimsElements()


jwt_token_get_responses = rest_responses.copy()
jwt_token_get_responses[HTTPOk.code] = JWTTokenGetResponse(
    description="Extract claims from provided token")


@jwt_token.get(content_type=('multipart/form-data', 'application/json'),
               schema=ClaimsObjectInfo(),
               validators=(check_cors_origin, colander_body_validator, set_cors_headers),
               response_schemas=jwt_token_get_responses)
def get_jwt_claims(request):
    """Extract claims from provided token"""
    sm = query_utility(ISecurityManager)  # pylint: disable=invalid-name
    if sm is None:
        return http_error(request, HTTPServiceUnavailable)
    configuration = IJWTSecurityConfiguration(sm)
    if not configuration.enabled:
        return http_error(request, HTTPServiceUnavailable)
    params = request.params if TEST_MODE else request.validated
    obj = params.get('obj')
    if configuration.proxy_mode:
        handler = IJWTProxyHandler(sm, None)
        if handler is not None:
            status_code, claims = handler.get_claims(request, obj)  # pylint: disable=assignment-from-no-return
            request.response.status_code = status_code
            return claims
    return get_request_claims(request, obj)


class JWTTokenPatchResponse(MappingSchema):
    """Token patch response"""

    body = TokensResult()


jwt_token_patch_responses = rest_responses.copy()
jwt_token_patch_responses[HTTPOk.code] = JWTTokenPatchResponse(
    description="Get new access token from valid refresh token")


@jwt_token.patch(require_csrf=False,
                 content_type=('multipart/form-data', 'application/json'),
                 jwt_object=REFRESH_OBJECT,
                 schema=ClaimsInfo(),
                 validators=(check_cors_origin, colander_body_validator, set_cors_headers),
                 response_schemas=jwt_token_patch_responses)
def refresh_jwt_token(request):
    """JWT token refresh service"""
    sm = query_utility(ISecurityManager)  # pylint: disable=invalid-name
    if sm is None:
        return http_error(request, HTTPServiceUnavailable)
    configuration = IJWTSecurityConfiguration(sm)
    if not configuration.enabled:
        return http_error(request, HTTPServiceUnavailable)
    # user remote authentication authority
    if configuration.proxy_mode:
        handler = IJWTProxyHandler(sm, None)
        if handler is not None:
            status_code, token = handler.refresh_token(request)  # pylint: disable=assignment-from-no-return
            request.response.status_code = status_code
            return token
    # refresh token locally
    claims = get_jwt_claims(request)
    if not claims:
        return http_error(request, HTTPForbidden)
    principal_id = claims.get('sub')
    if not principal_id:
        return http_error(request, HTTPUnauthorized)
    params = request.params if TEST_MODE else request.validated
    custom_claims = params.get('claims', {})
    result = {
        'status': STATUS.SUCCESS.value,
        configuration.access_token_name:
            create_jwt_token(request,
                             principal_id,
                             expiration=configuration.access_expiration,
                             obj=ACCESS_OBJECT,
                             **custom_claims)
    }
    if request.registry.settings.get('pyams_auth_jwt.auto_refresh', False):
        result[configuration.refresh_token_name] = create_jwt_token(request,
                                                                    principal_id,
                                                                    expiration=configuration.refresh_expiration,
                                                                    obj=REFRESH_OBJECT)
    return result


jwt_verify = Service(name=REST_VERIFY_ROUTE,
                     pyramid_route=REST_VERIFY_ROUTE,
                     description="JWT tokens verification")


@jwt_verify.options(validators=(check_cors_origin, set_cors_headers))
def jwt_verify_options(request):  # pylint: disable=unused-argument
    """JWT token verification OPTIONS handler"""
    return ''


class JWTVerifyGetResponse(MappingSchema):
    """Token verification response"""

    body = TokensResult()


jwt_verify_get_responses = rest_responses.copy()
jwt_verify_get_responses[HTTPOk.code] = JWTVerifyGetResponse(
    description="Get JWT token for authenticated principal")


@jwt_verify.get(schema=ClaimsObjectInfo(),
                validators=(check_cors_origin, set_cors_headers),
                response_schemas=jwt_verify_get_responses)
def get_current_jwt_token(request):
    """Get JWT token for authenticated principal"""
    identity = request.identity
    if identity is None:
        return http_error(request, HTTPUnauthorized)
    if Authenticated not in identity.principals:
        return http_error(request, HTTPForbidden)
    manager = query_utility(ISecurityManager)
    if manager is None:
        return http_error(request, HTTPServiceUnavailable)
    configuration = IJWTSecurityConfiguration(manager)
    if not configuration.enabled:
        return http_error(request, HTTPServiceUnavailable)
    custom_claims = request.params.get('claims', {})
    request.response.cache_expires(configuration.refresh_expiration)
    return {
        'status': STATUS.SUCCESS.value,
        configuration.access_token_name:
            create_jwt_token(request,
                             request.authenticated_userid,
                             expiration=configuration.access_expiration,
                             obj=ACCESS_OBJECT,
                             **custom_claims),
        configuration.refresh_token_name:
            create_jwt_token(request,
                             request.authenticated_userid,
                             expiration=configuration.refresh_expiration,
                             obj=REFRESH_OBJECT)
    }


class JWTVerifyPostResponse(MappingSchema):
    """JWT token verification response"""

    body = BaseStatusSchema()


jwt_verify_post_responses = rest_responses.copy()
jwt_verify_post_responses[HTTPOk.code] = JWTVerifyPostResponse(
    description="Verify JWT access token")


@jwt_verify.post(require_csrf=False,
                 jwt_object=ACCESS_OBJECT,
                 schema=ClaimsObjectInfo(),
                 validators=(check_cors_origin, colander_body_validator, set_cors_headers),
                 response_schemas=jwt_verify_post_responses)
def verify_jwt_token(request):
    """Verify JWT access token"""
    manager = query_utility(ISecurityManager)
    if manager is None:
        return http_error(request, HTTPServiceUnavailable)
    configuration = IJWTSecurityConfiguration(manager)
    if not configuration.enabled:
        return http_error(request, HTTPServiceUnavailable)
    claims = get_jwt_claims(request)
    if not claims:
        return http_error(request, HTTPForbidden)
    request.response.status_code = HTTPAccepted.code
    return {
        'status': STATUS.SUCCESS.value
    }



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS JWT authentication package.include module

This module is used for Pyramid integration.
"""

import jwt

from pyams_auth_jwt.interfaces import REST_TOKEN_PATH, REST_TOKEN_ROUTE, REST_VERIFY_PATH, REST_VERIFY_ROUTE
from pyams_auth_jwt.plugin import create_jwt_token, get_jwt_claims, jwt_object_view

__docformat__ = 'restructuredtext'


def include_package(config):
    """Pyramid package include"""

    # add translations
    config.add_translation_dirs('pyams_auth_jwt:locales')

    # add configuration directives
    config.add_request_method(create_jwt_token, 'create_jwt_token')
    config.add_request_method(get_jwt_claims, 'jwt_claims', reify=True)

    # add view deriver
    config.add_view_deriver(jwt_object_view, under='secured_view', over='owrapped_view')

    # register new REST API routes
    config.add_route(REST_TOKEN_ROUTE,
                     config.registry.settings.get(f'{REST_TOKEN_ROUTE}_route.path',
                                                  REST_TOKEN_PATH))
    config.add_route(REST_VERIFY_ROUTE,
                     config.registry.settings.get(f'{REST_VERIFY_ROUTE}_route.path',
                                                  REST_VERIFY_PATH))

    # update JWT algorithms
    try:
        import pycrypto  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        pass
    else:
        from jwt.algorithms import RSAAlgorithm  # pylint: disable=import-outside-toplevel
        jwt.unregister_algorithm('RS256')
        jwt.register_algorithm('RS256', RSAAlgorithm(RSAAlgorithm.SHA256))
        jwt.unregister_algorithm('RS512')
        jwt.register_algorithm('RS512', RSAAlgorithm(RSAAlgorithm.SHA512))

    try:
        import ecdsa  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        pass
    else:
        from jwt.algorithms import ECAlgorithm  # pylint: disable=import-outside-toplevel
        jwt.unregister_algorithm('ES256')
        jwt.register_algorithm('ES256', ECAlgorithm(ECAlgorithm.SHA256))
        jwt.unregister_algorithm('ES512')
        jwt.register_algorithm('ES512', ECAlgorithm(ECAlgorithm.SHA512))

    try:
        import pyams_zmi  # pylint: disable=import-outside-toplevel,unused-import
        config.scan()
    except ImportError:
        config.scan(ignore='pyams_auth_jwt.zmi')
